/*
  Copyright (C) 2023 GraphicsMagick Group

  This program is covered by multiple licenses, which are described in
  Copyright.txt. You should have received a copy of Copyright.txt with this
  package; otherwise see http://www.graphicsmagick.org/www/Copyright.html.

  Resource private methods.
*/

/*
  Define  declarations.
*/
#define MagickResourceInfinity ((magick_int64_t) (~((magick_uint64_t) 0) >> 1))

/*
  Method declarations.
*/
extern MagickPassFail
_UpdateMagickResourceHighwater(const ResourceType type,
                               const magick_int64_t highwater);
